using System.Collections.Generic;
using UnityEngine;
using System.Collections;

[ExecuteInEditMode]
public class Portal : MonoBehaviour {
	public Portal target;
	public Portal MyTarget{get{return target ?? this;}}
	
	private readonly Hashtable myCameras = new Hashtable();
	private Material mat;
	public float clipPlaneOffset = 0.0001f;
	private int oldRenderTextureSize;

    private const int RENDER_TEXTURE_SIZE = 1024;
	
	public bool disablePixelLights;
	
	public float scale = 10.0f;
	
	//create the mesh and stuff
	public void Awake () {
		Vector3[] vertices = new Vector3[4];
		Vector2[] uvs = new Vector2[4];
		int[] triangles = new int[6];
		
		uvs[0] = new Vector2(0,0);
		vertices[0] = new Vector3(-0.5f,-0.5f,0);
		uvs[1] = new Vector2(0,1);
		vertices[1] = new Vector3(-0.5f,0.5f,0);
		uvs[2] = new Vector2(1,1);
		vertices[2] = new Vector3(0.5f,0.5f,0);
		uvs[3] = new Vector2(1,0);
		vertices[3] = new Vector3(0.5f,-0.5f,0);
		
		triangles[0] = 2;
		triangles[1] = 1;
		triangles[2] = 0;
		
		triangles[3] = 0;
		triangles[4] = 3;
		triangles[5] = 2;
		
		Mesh mesh = new Mesh {vertices = vertices, triangles = triangles, uv = uvs};
	    mesh.RecalculateNormals();

        MeshFilter filter = gameObject.GetComponent<MeshFilter>()??gameObject.AddComponent<MeshFilter>();

	    filter.sharedMesh = mesh;

	    if(gameObject.GetComponent("MeshRenderer") == null)
			gameObject.AddComponent("MeshRenderer");
		
		
		if(mat==null){
			mat = new Material(Shader.Find("Portal Shader"));
		}
		renderer.sharedMaterial = mat;
		
		
	}

    public void Start()
    {
        OnEnable();
    }
	
	
	public void OnMouseDrag(){
		//transform.forward = Portal.RotateVecByQuat(transform.forward,Quaternion.Euler(0,5,0));
	}
	
	static Quaternion ScaleQuat(Quaternion q, float s){
		return new Quaternion(q.x*s,q.y*s,q.z*s,q.w*s);	
	}
	static Quaternion NormalizeQuat(Quaternion q){
		float f = Mathf.Sqrt(Mathf.Pow(q.x,2)+Mathf.Pow(q.y,2)+Mathf.Pow(q.z,2)+Mathf.Pow(q.w,2));
		return ScaleQuat(q,f);
	}
	
	public static Vector3 RotateVecByQuat(Vector3 v, Quaternion q){
	    Quaternion qV = new Quaternion(v.x,v.y,v.z,0);
		Quaternion qR = NormalizeQuat(q) * qV * Quaternion.Inverse(q);
		return new Vector3(qR.x,qR.y,qR.z);
	} 
	public bool IsFacing(Vector3 v,bool b){
		//if(b)
		//	Debug.Log((transform.InverseTransformPoint(v).z));
		return IsFacing(v);
	}
	public bool IsFacing(Vector3 v){ 
		return (transform.InverseTransformPoint(v).z)>0;
	}
	public float DistanceFrom(Vector3 v){
		return transform.InverseTransformPoint(v).z;
	}
	public void PortalTransform(Transform t,Transform o){
		o.position = PortalVector(t.position);
		o.rotation = Quaternion.LookRotation(PortalDir(t.forward),PortalDir(t.up));
		o.localScale = t.localScale * PortalScalar(1);
	}
	Vector3 GetLocalVector(Vector3 globalVector){
		return RotateVecByQuat(globalVector-transform.position,Quaternion.Inverse(transform.rotation))/scale;
	}
	
	Vector3 GetGlobalVector(Vector3 localVector){
		return transform.position+RotateVecByQuat(localVector*scale,transform.rotation);
	}
	
	public Quaternion PortalQuat(Quaternion q){
		Quaternion result = q * MyTarget.transform.rotation * Quaternion.Inverse(transform.rotation);
		//Vector3 euler = result.eulerAngles;
        //result.eulerAngles = new Vector3(-euler.x, -euler.y, -euler.z);
		//result *= Quaternion.Euler(new Vector3(0,180,0));
		//result.z *= -1;
		//result.y *= -1;
		//result.x *= -1;
		return result;
	}
	
	public Vector3 PortalVector(Vector3 v){
	    Vector3 localVec = transform.InverseTransformPoint(v);
		localVec.z *= MyTarget.scale / scale;
		localVec.x *= -1.0f;
		localVec.z *= -1.0f;
		return MyTarget.transform.TransformPoint(localVec);
//		return myTarget.GetGlobalVector(localVec);
	}
	
	public Vector3 PortalDir(Vector3 v){
		Vector3 localDir = transform.InverseTransformDirection(v);
		
		localDir.x *= -1.0f;
		localDir.z *= -1.0f;
		return MyTarget.transform.TransformDirection(localDir);
	}
	
	public Vector3 PortalVectorNonRotated(Vector3 v){
		return MyTarget.GetGlobalVector(GetLocalVector(v));
	}
	
	public Quaternion Opposite(Quaternion q){
		Quaternion upQuat = new Quaternion(0,1,0,0);
		Quaternion forwardQuat = new Quaternion(0,0,1,0);
		//Quaternion leftQuat = new Quaternion(1,0,0,0);
		Quaternion result = q;
		result = result * upQuat;
		result = result * forwardQuat;
		return result;
	}
	
	public void OnDrawGizmos(){
		//if(IsFacing(Camera.main.transform.position))
		//	Gizmos.DrawSphere(transform.position,0.5f);
		/*
		if(this.name=="A"){
			GameObject g = new GameObject("test");
			Transform t = g.transform;
			
			t.position = Camera.main.transform.position;
			t.rotation = Camera.main.transform.rotation;
			
			
			Gizmos.DrawSphere(t.position,0.5f);
			Gizmos.DrawLine(t.position,t.position+t.forward*2);
			Gizmos.DrawLine(t.position,t.position+t.up);
			
			t.position = PortalVector(t.position);
			t.rotation = Quaternion.LookRotation(PortalDir(t.forward),PortalDir(t.up));
			
			//Debug.Log(t.forward);
			Gizmos.DrawSphere(t.position,0.5f);
			Gizmos.DrawLine(t.position,t.position+t.forward*2);
			Gizmos.DrawLine(t.position,t.position+t.up);
			
			g.hideFlags = HideFlags.HideAndDontSave;
		}
		*/
		//Gizmos.DrawLine(transform.position,transform.position+transform.forward);
		//Debug.Log((Quaternion.LookRotation(-transform.forward).eulerAngles-Quaternion.LookRotation(transform.forward).eulerAngles));
		//Gizmos.DrawLine(transform.position,transform.position+RotateVecByQuat(new Vector3(0,0,1) * 2,transform.rotation));
		//Vector3 euler = transform.rotation.eulerAngles;
		//Quaternion portaledQuat = Opposite(transform.rotation);
		//Transform t = new Transform();
		//t.rotation = transform.rotation;
		//portaledQuat =
		//Gizmos.DrawLine(transform.position,transform.position+RotateVecByQuat(new Vector3(0,0,1) * 2,portaledQuat));
		/*
		Gizmos.DrawSphere(Vector3.zero,0.5f);
		
		Gizmos.DrawLine(myTarget.transform.position,myTarget.transform.position+RotateVecByQuat(new Vector3(0,0,5),portaledQuat));
		*/
	}

    private readonly List<Camera> renderQueue = new List<Camera>();
    private List<RenderTexture> textures;
    private List<Matrix4x4> matrices;

    private const int MAX_DEPTH = 10;
    //private readonly Hashtable textures = new Hashtable();
    //private readonly Hashtable matrices = new Hashtable();
 

    public void OnEnable()
    {
        textures = new List<RenderTexture>();
        for (int i = 0; i < MAX_DEPTH+5; i++)
        {
            int texSize = (int) Mathf.Max(128, RENDER_TEXTURE_SIZE/Mathf.Pow(2, Mathf.FloorToInt(i/2.0f)));
            textures.Add(new RenderTexture(texSize,texSize, 16)
            {
                name = "_PortalRender" + GetInstanceID() + Random.value,
                hideFlags = HideFlags.HideInHierarchy,
                isPowerOfTwo = true
            });
        }

        matrices = new List<Matrix4x4>();
        for (int i = 0; i < MAX_DEPTH+5; i++)
        {
            matrices.Add(new Matrix4x4());
        }
    }

	public void OnRenderObject()
    {
        //if (renderCount == 0){
            //if (RenderTexture.active == renderTexture)
            //    RenderTexture.active = null;

		    //DestroyImmediate(renderTexture);
        //}
        //Debug.Log("onrenderend"+" "+this);
        Camera cam = Camera.current;

        if (!cam)
            return;

	    renderQueue.Remove(cam);

	    if (renderQueue.Count == 0) return;

	    if(renderQueue[renderQueue.Count-1]!=null){
	        if (mat.HasProperty("_ReflectionTex"))
                mat.SetTexture("_ReflectionTex", textures[renderQueue.Count - 1]);

	        mat.SetMatrix("_ProjMatrix", matrices[renderQueue.Count - 1]);
	    }

    }

	public void OnWillRenderObject()
	{
		//Debug.Log("onrenderstart"+" "+this);
        if (!enabled || !renderer || !renderer.sharedMaterial || !renderer.enabled || !SystemInfo.supportsRenderTextures)
			return;
		
		Camera cam = Camera.current;
		
        if( !cam )
            return;

        renderQueue.Add(cam);
        //print(name + " start " + renderQueue.Count);
        //Debug.Log(this + " " + Vector3.Dot(transform.forward, (transform.position - cam.transform.position).normalized));
		//Debug.Log(this+" "+cam);
        //if (Vector3.Dot(MyTarget.transform.forward, PortalDir(cam.transform.forward)) < 0)
        if (Vector3.Dot(transform.forward, (transform.position - cam.transform.position)) > 0){
            //Debug.Log(Vector3.Dot(transform.forward, (transform.position - cam.transform.position)));
			return;
        }
		
		//Debug.DrawLine(cam.transform.position,transform.position);
		//if( insideRendering )
        //    return;
		
//        insideRendering = true;
		
		if(mat==null){
			mat = new Material(Shader.Find("Portal Shader"));
		}

        renderer.sharedMaterial = mat;
        //if (_renderCount > 1)
            if(cam.GetComponent<DuplicateCamera>())
            {
                if (cam.GetComponent<DuplicateCamera>().level >= MAX_DEPTH)
                {
                    if (mat.HasProperty("_ReflectionTex"))
                        mat.SetTexture("_ReflectionTex", (Texture2D)Resources.Load("Red"));
                    return;
                }
            }
            //if (mat.HasProperty("_ReflectionTex"))
             //   mat.SetTexture("_ReflectionTex", (Texture2D)Resources.Load("Red"));
            //if(renderCount>0)
            //return;
            //return;


	    //renderCount++;
        //_renderCount++;
//		UnityEditor.SceneView.GetAllSceneCameras()[0]
		Camera portalCamera;
		CreateCamera(cam,out portalCamera);
		
		UpdateCameraModes(cam,portalCamera);
		portalCamera.transform.position = PortalVector(cam.transform.position);
		Vector3 newForward = PortalDir(cam.transform.forward);
		Vector3 newUp = PortalDir(cam.transform.up);
		portalCamera.transform.rotation = Quaternion.LookRotation(newForward,newUp);
		
		Vector4 clipPlane = CameraSpacePlane( portalCamera, MyTarget.transform.position, MyTarget.transform.forward, 1.0f );
		Matrix4x4 projection = cam.projectionMatrix;
        CalculateObliqueMatrix (ref projection, clipPlane);

        portalCamera.projectionMatrix = projection;
        
        RenderTexture renderTexture = textures[renderQueue.Count-1];
		portalCamera.targetTexture = renderTexture;

	    RenderTexture prev = RenderTexture.active;
        RenderTexture.active = renderTexture;
		portalCamera.Render();
        RenderTexture.active = prev;

	    //textures[cam] = renderTexture;

        if (mat.HasProperty("_ReflectionTex"))
            mat.SetTexture("_ReflectionTex", renderTexture);
	    //portalCamera.projectionMatrix = oldProjection;
		
		Matrix4x4 scaleOffset = Matrix4x4.TRS(
            (new Vector3(0.5f,0.5f,0.5f)), Quaternion.identity, (new Vector3(0.5f,0.5f,0.5f)));
		Vector3 lScale = transform.lossyScale;
        Matrix4x4 mtx = transform.localToWorldMatrix * Matrix4x4.Scale( (new Vector3(1.0f/lScale.x, 1.0f/lScale.y, 1.0f/lScale.z)) ); 
        mtx = scaleOffset * cam.projectionMatrix * cam.worldToCameraMatrix * mtx;

	    matrices[renderQueue.Count-1] = mtx;
        mat.SetMatrix( "_ProjMatrix", mtx );

        //_renderCount--;// = false;
        //renderCount--;

    }
	private float oldScale;
    //rescale the object based on the scale
	public void Update() {
	    if (oldScale == scale) return;
	    if(scale==0)
	        scale = 0.01f;
	    transform.localScale = new Vector3(0.5f,1,0)*scale + new Vector3(0,0,1);
	    foreach(Transform c in transform.GetComponentInChildren<Transform>()){
	        c.localScale = new Vector3(c.localScale.x,c.localScale.y,scale/10.0f);
	    }
	    oldScale = scale;
	}
	
	// modifies projection matrix in place
	// clipPlane is in camera space
    static void CalculateObliqueMatrix( ref Matrix4x4 projection, Vector4 clipPlane )
	{
        Vector4 q = projection.inverse * new Vector4(
                                             Sgn(clipPlane.x),
                                             Sgn(clipPlane.y),
                                             1.0f,
                                             1.0f
                                             );
		
	    Vector4 c = clipPlane * (2.0F / (Vector4.Dot (clipPlane, q)));
		
		projection.SetRow(2,c-projection.GetRow(3));
	}
	
	private static float Sgn(float a)
    {
        if (a > 0.0f) return 1.0f;
        if (a < 0.0f) return -1.0f;
        return 0.0f;
    }
	
	private Vector4 CameraSpacePlane (Camera cam, Vector3 pos, Vector3 normal, float sideSign)
    {
        Vector3 offsetPos = pos + normal * clipPlaneOffset;
        Matrix4x4 m = cam.worldToCameraMatrix;
        Vector3 cpos = m.MultiplyPoint( offsetPos );
        Vector3 cnormal = m.MultiplyVector( normal ).normalized * sideSign;
        return new Vector4( cnormal.x, cnormal.y, cnormal.z, -Vector3.Dot(cpos,cnormal) );
    }

	
	private static void UpdateCameraModes( Camera src, Camera dest )
    {
        if( dest == null )
            return;
        // set camera to clear the same way as current camera
        dest.clearFlags = src.clearFlags;
        dest.backgroundColor = src.backgroundColor;        
        if( src.clearFlags == CameraClearFlags.Skybox )
        {
            Skybox sky = src.GetComponent(typeof(Skybox)) as Skybox;
            Skybox mysky = dest.GetComponent(typeof(Skybox)) as Skybox;
            if(mysky){
                if( !sky || !sky.material )
                {
                    mysky.enabled = false;
                }
                else
                {
                    mysky.enabled = true;
                    mysky.material = sky.material;
                }
            }
        }
        // update other values to match current camera.
        // even if we are supplying custom camera&projection matrices,
        // some of values are used elsewhere (e.g. skybox uses far plane)
        dest.farClipPlane = src.farClipPlane;
        dest.nearClipPlane = src.nearClipPlane;
        dest.orthographic = src.orthographic;
        dest.fieldOfView = src.fieldOfView;
        dest.aspect = src.aspect;
        dest.orthographicSize = src.orthographicSize;
    }
	public float PortalScalar(float f){
		return f * MyTarget.scale/scale;
	}
	public void OnDisable()
    {
		//Debug.Log(this); 
		//if(renderTexture){
		//	DestroyImmediate(renderTexture);
		//	renderTexture = null;
		//}
		foreach( DictionaryEntry e in myCameras )
            DestroyImmediate( ((Camera)e.Value).gameObject );

	    foreach (RenderTexture e in textures)
	    {
	        DestroyImmediate(e);
	    }

		
		//renderer.material = null;
		
		//if(mat!=null)
		//	DestroyImmediate(mat);
		
		//mat = null;
		
        textures.Clear();
        myCameras.Clear();
        matrices.Clear();
	}
    //void ResetRenderTexture(){
    //    if(renderTexture){
    //        if (RenderTexture.active == renderTexture)
    //            RenderTexture.active = null;

    //        DestroyImmediate(renderTexture);
    //    }
		
    //    renderTexture = new RenderTexture(renderTextureSize,renderTextureSize,16) {
    //                                                                                  name = "_PortalRender" + GetInstanceID(),
    //                                                                                  hideFlags = HideFlags.HideAndDontSave,
    //                                                                                  isPowerOfTwo = true
    //                                                                              };
    //}


	void CreateCamera(Camera currentCamera, out Camera newCamera){
		newCamera = myCameras[currentCamera] as Camera;
		if(!newCamera){
			GameObject g = new GameObject("Portal Camera id"+name+" for "+currentCamera.name,typeof(Camera),typeof(Skybox));
		    DuplicateCamera dup = g.AddComponent<DuplicateCamera>();
            if (currentCamera.GetComponent<DuplicateCamera>())
                dup.level = currentCamera.GetComponent<DuplicateCamera>().level + 1;
		
			newCamera = g.camera;
			newCamera.enabled = false;
			newCamera.transform.position = MyTarget.transform.position;
			newCamera.transform.rotation = MyTarget.transform.rotation;
			newCamera.gameObject.AddComponent("FlareLayer");
			g.hideFlags = HideFlags.HideAndDontSave;
			myCameras[currentCamera] = newCamera;
		}
			
	}
}
	