using UnityEngine;
using System.Collections.Generic;

public class PortalableObject : MonoBehaviour
{
	Vector3 lastPos;
	PortalableObject clone;
	bool isClone;
	Portal currentPortal;
	PortalableObject owner;
	public PortalableObject Other{ get{
			if(isClone) 
				return owner; 
			if(clone)
				return clone;
			return null;
		}}
	
	bool hadGravity;
	
	Vector3 oldPortalPos;
	Quaternion oldPortalRot;
	
	public void Awake(){
		currentPortal = null;
		if(!renderer.material||(!renderer.material.HasProperty("_planePos")))
			renderer.material = Resources.Load("DefaultPortalMaterial") as Material;
		if(!rigidbody){
			gameObject.AddComponent("Rigidbody");	
		}
		rigidbody.collisionDetectionMode = CollisionDetectionMode.ContinuousDynamic;
		hadGravity = rigidbody.useGravity;
		//rigidbody.useGravity = false;
	}
	// Use this for initialization
	public void Start ()
	{
		lastPos = transform.position;
		//rigidbody.angularVelocity=new Vector3(0,0,0.5f);
	}

	// Update is called once per frame
    public void Update()
	{
		if(isClone)
			return;
		
		if(currentPortal){
			InPortal(currentPortal);
			if(collisions.Count>0||clone.collisions.Count>0){
				SyncCollisions();
			}
			rigidbody.useGravity = false;
			
			if(currentPortal.transform.position != oldPortalPos||!currentPortal.transform.rotation.Equals(oldPortalRot)){
				PortalObject(currentPortal,clone);
			}
			oldPortalPos = currentPortal.transform.position;
			oldPortalRot = currentPortal.transform.rotation;
		}
		else{
			if(hadGravity)
				rigidbody.useGravity = true;
			foreach(RaycastHit hit in Physics.RaycastAll(lastPos,(transform.position-lastPos),Vector3.Distance(lastPos,transform.position))){
				if(hit.collider.gameObject.GetComponent("Portal")){
					Portal p = hit.collider.gameObject.GetComponent("Portal") as Portal;
					if(p.IsFacing(lastPos)&&!p.IsFacing(transform.position)){
						PortalObject(p,this);
						//currentPortal = p.myTarget;
					}
					//InPortal(currentPortal);
				}
			}
		}
		
		if(clone&&!currentPortal){
			Destroy(clone.gameObject);
			clone = null;
		}
		
		//UpdateSynchronously();
		
		
		
		lastPos = transform.position;
		if(!clone){
			return;
		}
		
		
	}
	void PortalObject(Portal p, PortalableObject o){
		o.rigidbody.velocity = p.PortalDir(rigidbody.velocity)*p.PortalScalar(1);
		p.PortalTransform(transform,o.transform);
		o.rigidbody.angularVelocity = p.PortalDir(rigidbody.angularVelocity);
		o.currentPortal = p.MyTarget;
	}
    public void OnWillRenderObject()
    {
		if(currentPortal){
			if(renderer.material.HasProperty("_planePos")){
				renderer.material.SetVector("_planePos",currentPortal.transform.position);
				renderer.material.SetVector("_planeNorm",-currentPortal.transform.forward);
			}
		}
		else{
			//Debug.Log(this);
			renderer.material.SetVector("_planePos",Vector3.zero);
			renderer.material.SetVector("_planeNorm",Vector3.zero);
		}
	}
	void InPortal(Portal p){
		if(isClone)
			return;
		if(p.IsFacing(lastPos)&&!p.IsFacing(transform.position)){
			PortalObject(p,this);
			/*
			p.PortalTransform(transform,transform);
			if(rigidbody){
				if(!rigidbody.isKinematic)
					rigidbody.velocity = p.PortalDir(rigidbody.velocity) * p.PortalScalar(1);
				rigidbody.angularVelocity = p.PortalDir(rigidbody.angularVelocity);
			}
			*/
			currentPortal = p.MyTarget;
			//Debug.Log(GameManager.selectedObject+" "+this.gameObject);
			
			if(clone){
				PortalObject(currentPortal,clone);	
			
				if(GameManager.selectedObject){
					if(GameManager.selectedObject.Equals(gameObject))
						GameManager.selectedObject = clone.gameObject;
					else if(GameManager.selectedObject.Equals(clone.gameObject))
						GameManager.selectedObject = gameObject;
				}
			}
			/*
			if(clone){
				clone.rigidbody.velocity = currentPortal.PortalDir(rigidbody.velocity)*currentPortal.PortalScalar(1);
				currentPortal.PortalTransform(transform,clone.transform);
			}
			*/
		}
		else{
			if(!clone){
				clone = Instantiate(this) as PortalableObject;
                //Physics.IgnoreCollision(clone.collider,collider);
				clone.isClone = true;
				clone.owner = this;
				clone.rigidbody.useGravity = false;
				//clone.currentPortal = currentPortal.myTarget;
				clone.renderer.material.SetVector("_planePos",currentPortal.MyTarget.transform.position);
				clone.renderer.material.SetVector("_planeNorm",-currentPortal.MyTarget.transform.forward);
				PortalObject(currentPortal,clone);
			}
		}
		if(clone){
			clone.currentPortal = currentPortal.MyTarget;
			
			/*
			rigidbody.useGravity=false;
			clone.rigidbody.useGravity=false;
			rigidbody.AddForce(Physics.gravity+currentPortal.myTarget.PortalDir(Physics.gravity));
			*/
			/*
			//clone.currentPortal = currentPortal.myTarget;
			float weight = Mathf.Clamp(0.5f+currentPortal.DistanceFrom(transform.position)/(collider.bounds.size.magnitude/2),0,1);
			//weight = 0.5f;
			float cloneWeight = 1-weight;
			Vector3 middleVelocity = (rigidbody.velocity*weight + currentPortal.myTarget.PortalDir(clone.rigidbody.velocity) * currentPortal.myTarget.PortalScalar(1)*cloneWeight)/1;
			Vector3 middlePosition = (transform.position*weight + currentPortal.myTarget.PortalVector(clone.transform.position)*cloneWeight)/1;
			
			rigidbody.velocity = middleVelocity;
			transform.position = middlePosition;
			*/
			//rigidbody.AddForce(0.5f*(Physics.gravity+currentPortal.myTarget.PortalDir(Physics.gravity)));
			
			//clone.rigidbody.AddForce(0.5f*(Physics.gravity+currentPortal.PortalDir(Physics.gravity)));
			/*
			if(!clone.rigidbody.isKinematic)
				clone.rigidbody.velocity = currentPortal.PortalDir(rigidbody.velocity)*currentPortal.PortalScalar(1);
			currentPortal.PortalTransform(transform,clone.transform);
			*/
			//PortalObject(currentPortal,clone);
			/**/
		}
			
		/*
		if(myTurn){
			//clone just acted!
			
			//for next frame
			
			//rigidbody.WakeUp();
			//clone.rigidbody.isKinematic = true;
			//clone.rigidbody.Sleep();
			clone.rigidbody.WakeUp();
			clone.rigidbody.velocity = currentPortal.PortalDir(rigidbody.velocity)*currentPortal.PortalScalar(1);
			currentPortal.PortalTransform(transform,clone.transform);
			rigidbody.Sleep();
		}
		else{
			//clone.rigidbody.velocity = currentPortal.PortalDir(rigidbody.velocity)*currentPortal.PortalScalar(1);
			//we just acted!
			rigidbody.WakeUp();
			currentPortal.myTarget.PortalTransform(clone.transform,transform);
			rigidbody.velocity = currentPortal.myTarget.PortalDir(clone.rigidbody.velocity)*currentPortal.myTarget.PortalScalar(1);
			
			clone.rigidbody.Sleep();
			//for next frame
			//clone.rigidbody.WakeUp();
			
			//rigidbody.Sleep();
		}
		myTurn = !myTurn;
		*/
	}

    readonly List<Collision> collisions = new List<Collision>();
	//Vector3 oldVelocity = Vector3.zero;
    public void FixedUpdate()
    {
		if(currentPortal){
			if(!clone)
				return;
			float weight = Mathf.Clamp(0.5f+currentPortal.DistanceFrom(transform.position)/(collider.bounds.size.magnitude/2),0,1);
			//weight = 0.5f;
			float otherWeight = 1 - weight;
			if(!isClone){
				
				if(collisions.Count>0||clone.collisions.Count>0){
					SyncCollisions();
				}
				//rigidbody.AddForce(Physics.gravity*0.5f+currentPortal.myTarget.PortalDir(Physics.gravity)*0.5f);
				//clone.rigidbody.AddForce(Physics.gravity*0.5f+currentPortal.PortalDir(Physics.gravity)*0.5f);
				Vector3 myGrav = (Physics.gravity*weight+currentPortal.MyTarget.PortalDir(Physics.gravity)*otherWeight)*currentPortal.MyTarget.PortalScalar(1);
				
				//Debug.Log(myGrav+" "+Physics.gravity);
				//myGrav/=Time.deltaTime;
				rigidbody.AddForce(myGrav);
				//currentPortal.PortalDir(myGrav);
				clone.rigidbody.AddForce(currentPortal.PortalDir(myGrav)*currentPortal.PortalScalar(1));
			}
		}
	}
	public void SetVelocity(Vector3 newVel){
		rigidbody.velocity = newVel;
		if(currentPortal)
			Other.rigidbody.velocity = currentPortal.PortalDir(newVel)*currentPortal.PortalScalar(1);
	}
	void SyncCollisions(){
		if(!clone)
			return;
		if(collisions.Count>0){
			if(clone.collisions.Count>0){
				
				Vector3 newVelocity = (rigidbody.velocity+currentPortal.MyTarget.PortalDir(clone.rigidbody.velocity)*currentPortal.MyTarget.PortalScalar(1))/2;
				rigidbody.velocity = newVelocity;
				clone.rigidbody.velocity = currentPortal.PortalDir(newVelocity)*currentPortal.PortalScalar(1);
				
				Vector3 newAngularVelocity = (rigidbody.angularVelocity+currentPortal.MyTarget.PortalDir(clone.rigidbody.angularVelocity)*currentPortal.MyTarget.PortalScalar(1))/2;
				rigidbody.angularVelocity = newAngularVelocity;
				clone.rigidbody.angularVelocity = currentPortal.PortalDir(newAngularVelocity)*currentPortal.PortalScalar(1);
				
				const float weight = 0.5f;
				const float otherWeight = 1- weight;
				Vector3 newPosition = (transform.position*weight+currentPortal.MyTarget.PortalVector(clone.transform.position)*otherWeight);
				transform.position = newPosition;
				clone.transform.position = currentPortal.PortalVector(newPosition);
				
				Vector3 portaledForward = currentPortal.MyTarget.PortalDir(clone.transform.forward);
				Vector3 portaledUp = currentPortal.MyTarget.PortalDir(clone.transform.up);
				
				Quaternion newRotation = Quaternion.Lerp(transform.rotation,Quaternion.LookRotation(portaledForward,portaledUp),weight);
				clone.transform.rotation = Quaternion.Lerp(clone.transform.rotation,Quaternion.LookRotation(currentPortal.PortalDir(transform.forward),currentPortal.PortalDir(transform.up)),otherWeight);
				transform.rotation = newRotation;
				
				/*
				p.PortalTransform(transform,o.transform);
				o.rigidbody.angularVelocity = p.PortalDir(rigidbody.angularVelocity);
				o.currentPortal = p.myTarget;
				*/
			}
			else
				PortalObject(currentPortal,clone);
		}
		else
			clone.PortalObject(currentPortal.MyTarget,this);
		
	}
    public void OnCollisionStay(Collision coll)
    {
		if(currentPortal){
			collisions.Add(coll);
		}
	}
    public void OnCollisionEnter(Collision coll)
    {
		if(currentPortal){
			collisions.Add(coll);
		}
		//}
	}
    public void OnCollisionExit(Collision coll)
    {
		if(currentPortal){
			collisions.Add(coll);
		}
		//}
	}
    public void OnTriggerEnter(Collider other)
    {
		if(isClone)
			return;
		if(other.gameObject.GetComponent("Portal")){
			if(currentPortal!=other.gameObject.GetComponent("Portal") as Portal){
				oldPortalPos = (other.gameObject.GetComponent("Portal") as Portal).transform.position;
				oldPortalRot = (other.gameObject.GetComponent("Portal") as Portal).transform.rotation;	
			}
			currentPortal = other.gameObject.GetComponent("Portal") as Portal;
		}
    }

    public void OnTriggerStay(Collider other)
    {
		if(isClone)
			return;
		if(other.gameObject.GetComponent("Portal")){
			if(currentPortal!=other.gameObject.GetComponent("Portal") as Portal){
				oldPortalPos = (other.gameObject.GetComponent("Portal") as Portal).transform.position;
				oldPortalRot = (other.gameObject.GetComponent("Portal") as Portal).transform.rotation;	
			}
			currentPortal = other.gameObject.GetComponent("Portal") as Portal;
		}
	}

    public void OnTriggerExit(Collider other)
    {
		if(isClone)
			return;
		if(other.gameObject.GetComponent("Portal")){
			if((other.gameObject.GetComponent("Portal") as Portal)==currentPortal){
				currentPortal = null;
			}
		}
	}
}

